import { world, system, Entity } from '@minecraft/server';
import table from "./table";
import profanityFilter from './profanity-filter';

const BUBBLES = {};
const MAX_TIME = 60;// 3 seconds

Entity.prototype.rideable = function () { return this.getComponent("minecraft:rideable"); }

function loadBubbles(e, add) {
    if (add) {
        if (e.typeId === "bubble:chat") BUBBLES[e.id] = e;
    } else delete BUBBLES[e];
}

function setuo() {
    world.getDimension('overworld').getEntities().forEach(e => {
        if (e.typeId === "bubble:chat") e.remove();
    });
    update();
}

function update() {
    for (let i in BUBBLES) {
        let bubble = BUBBLES[i];
        if (!(bubble && bubble.isValid()) && !bubble.hasTag("bubble:set")) continue;
        if (bubble.timer != undefined) {
            if (system.currentTick - bubble.timer >= MAX_TIME) {
                bubble.remove();
                continue;
            }
        }
        try {
            let parent = world.getEntity(`${bubble.getDynamicProperty("parent") || ""}`);
            if (parent && parent.isValid()) {
                bubble.setRotation(parent.getRotation());
                parent.rideable().getRiders().forEach((e, i) => {
                    if (e.typeId === "bubble:chat") bubble.getComponent("minecraft:skin_id").value = i;
                });
            } else bubble.remove();

            if (bubble.getDynamicProperty("bubble:targets")) {
                let players = JSON.parse(bubble.getDynamicProperty("bubble:targets"));
                if (players.length > 1) bubble.playAnimation("animation.bubble.show", { stopExpression: "q.is_alive == 0", players });
                else bubble.playAnimation("animation.bubble.show", { stopExpression: "q.is_alive == 0" });
                bubble.getComponent("minecraft:mark_variant").value = Math.max(players.length-1,0);
            }
        } catch (err) {
            bubble.remove();
            throw err;
        }
    }

    system.run(update);
}

world.beforeEvents.chatSend.subscribe(e => {
    let { message, sender } = e, target = "";
    if (message.startsWith("@p")) {
        const match = message.match(/\[(.*?)\]/);
        if (match) {
            let tmp = message.replace(match[0], "");
            let firstSpace = tmp.indexOf(" ");
            target = tmp.substring(0, firstSpace) + match[0];
            message = tmp.substring(firstSpace + 1);
        }
        else {
            let firstSpace = message.indexOf(" ");
            target = message.substring(0, firstSpace);
            message = message.substring(firstSpace + 1);
        }
    } else {
        const match = message.match(/\[(.*?)\]/);
        if (match) {
            let tmp = message.replace(match[0], "");
            target = match[1];
            message = tmp.substring(tmp.indexOf(" ") + 1);
        }
        else message = e.message;
    }
    sender.runCommandAsync(`tell ${target} ${message}`).then(res => {
        if (res.successCount == 0)
            sender.runCommandAsync(`say ${message}`)
                .then(_ => createBubble(sender, message))
                .catch(err => sender.sendMessage(`\u00A7c${err}\u00A7r`));
        else {
            let id = `id_${system.currentTick}`;
            sender.runCommandAsync(`tag ${target} add ${id}`)
                .then(_ => createBubble(sender, message, id));
        }
    }).catch(_ => {
        sender.runCommandAsync(`say ${message}`)
            .then(_ => createBubble(sender, message))
            .catch(err => sender.sendMessage(`\u00A7c${err}\u00A7r`));
    });
    e.cancel = true;
});

function createBubble(sender, message, id) {
    let players = [sender.nameTag], bubble;
    if (id) {
        for (let player of world.getAllPlayers()) {
            if (player.hasTag(id)) players.push(player.nameTag);
            player.removeTag(id);
        }
    }
    message = profanityFilter.censor(message);
    sender.rideable().getRiders().forEach(e => {
        if (e.typeId === "bubble:chat") {
            if (bubble) e.remove();
            else bubble = e;
        }
    });
    if (bubble && bubble.isValid()) applyCharacters(bubble);
    else {
        const b = sender.rideable().getRiders()[1];
        if (b) sender.rideable().ejectRider(b);
        sender.runCommandAsync("ride @s summon_rider bubble:chat").then(_ => {
            sender.rideable().getRiders().some((e, i) => {
                if (e.typeId === "bubble:chat") {
                    e.setDynamicProperty("parent", sender.id);
                    e.getComponent("minecraft:skin_id").value = i;
                    e.addTag("bubble:set");
                    applyCharacters(e);
                    return true;
                }
            });
        });
    }

    function applyCharacters(e) {
        let msg = message.split(""), index = 0;
        for (let i = 0; i < 32; i++) {
            e.resetProperty(`bubble:char${i}`);
        }
        for (let i = 1; i <= msg.length; i += 2) {
            if (index > 31) break;
            let value = 0;
            if (msg[i - 1]) value = table(msg[i - 1]) * 1e2;
            if (msg[i]) value += table(msg[i]);
            e.setProperty(`bubble:char${index++}`, value);
        }
        e.playAnimation("animation.bubble.hide", { stopExpression: "q.is_alive == 0" });
        e.timer = system.currentTick;
        e.setDynamicProperty("bubble:targets", JSON.stringify(players));
    }
}

world.afterEvents.worldInitialize.subscribe(setuo);
world.afterEvents.entityLoad.subscribe(e => loadBubbles(e.entity, true));
world.afterEvents.entitySpawn.subscribe(e => loadBubbles(e.entity, true));
world.afterEvents.entityDie.subscribe(e => loadBubbles(e.deadEntity.id));
world.beforeEvents.entityRemove.subscribe(e => loadBubbles(e.removedEntity.id));